from collections.abc import Callable

from py_flare_common._hexstr.hexstr import to_bytes

from .byte_parser import ByteParser, ParseError
from .types import (
    FdcSubmit1,
    FdcSubmit2,
    FtsoSubmit1,
    FtsoSubmit2,
    ParsedMessage,
    ParsedPayload,
    Signature,
    SubmitSignatures,
    SubmitSignaturesMessage,
    T,
    U,
)

__all__ = [
    "parse_generic_tx",
    "parse_submit1_tx",
    "parse_submit2_tx",
    "parse_submit_signature_tx",
]


def _default_parse(b: bytes) -> bytes:
    return b


def parse_generic_tx(
    message: bytes | str,
    pid_100_parse: Callable[[bytes], T] = _default_parse,
    pid_200_parse: Callable[[bytes], U] = _default_parse,
) -> ParsedMessage[T, U]:
    kwargs: dict[str, ParsedPayload | None] = {"ftso": None, "fdc": None}
    message = to_bytes(message)
    bp = ByteParser(message)

    while not bp.is_empty():
        protocol_id = bp.uint8()
        voting_round_id = bp.uint32()
        payload_length = bp.uint16()
        payload = bp.next_n(payload_length)

        if protocol_id == 100:
            parsed = pid_100_parse(payload)
            kwargs["ftso"] = ParsedPayload(
                protocol_id, voting_round_id, payload_length, parsed
            )

        if protocol_id == 200:
            parsed = pid_200_parse(payload)
            kwargs["fdc"] = ParsedPayload(
                protocol_id, voting_round_id, payload_length, parsed
            )

    return ParsedMessage(**kwargs)


def parse_submit1_tx(message: bytes | str) -> ParsedMessage[FtsoSubmit1, FdcSubmit1]:
    return parse_generic_tx(message, ftso_submit1, fdc_submit1)


def parse_submit2_tx(message: bytes | str) -> ParsedMessage[FtsoSubmit2, FdcSubmit2]:
    return parse_generic_tx(message, ftso_submit2, fdc_submit2)


def parse_submit_signature_tx(
    message: bytes | str,
) -> ParsedMessage[SubmitSignatures, SubmitSignatures]:
    return parse_generic_tx(message, submit_signatures, submit_signatures)


def ftso_submit1(payload: bytes) -> FtsoSubmit1:
    if len(payload) != 32:
        raise ParseError("Invalid payload length: expected 32 bytes.")
    return FtsoSubmit1(payload)


def fdc_submit1(payload: bytes) -> FdcSubmit1:
    if payload:
        raise ParseError("Invalid payload length: expected 0 bytes.")
    return FdcSubmit1()


def ftso_submit2(payload: bytes) -> FtsoSubmit2:
    EMPTY_FEED_VALUE = "0" * 8

    bp = ByteParser(payload)
    random = bp.uint256()
    values: list[int | None] = []

    while not bp.is_empty():
        raw_value = bp.next_n(4).hex()
        value = int(raw_value, 16) - 2**31 if raw_value != EMPTY_FEED_VALUE else None
        values.append(value)

    return FtsoSubmit2(random=random, values=values)


def fdc_submit2(payload: bytes) -> FdcSubmit2:
    bp = ByteParser(payload)
    n_requests = bp.uint16()

    votes = bp.drain()
    bit_vector = [False for _ in range(n_requests)]

    for j, byte in enumerate(reversed(votes)):
        for shift in range(8):
            i = n_requests - 1 - j * 8 - shift
            if i < 0 and (byte >> shift) & 1 == 1:
                raise ParseError("Invalid payload length.")
            elif i >= 0:
                bit_vector[i] = (byte >> shift) & 1 == 1

    return FdcSubmit2(
        number_of_requests=n_requests,
        bit_vector=bit_vector,
    )


def submit_signatures_type_0(payload: bytes) -> SubmitSignatures:
    payload_bp = ByteParser(payload)
    message_to_parse = payload_bp.next_n(38)
    signature_to_parse = payload_bp.next_n(65)
    unsignedMessage = payload_bp.drain()

    message_bp = ByteParser(message_to_parse)
    protocol_id = message_bp.uint8()
    message_bp.next_n(4)
    random_quality_score = message_bp.uint8()
    merkle_root = message_bp.drain().hex()

    message = SubmitSignaturesMessage(
        protocol_id=protocol_id,
        random_quality_score=random_quality_score,
        merkle_root=merkle_root,
    )

    signature_bp = ByteParser(signature_to_parse)
    v = signature_bp.next_n(1).hex()
    r = signature_bp.next_n(32).hex()
    s = signature_bp.next_n(32).hex()

    signature = Signature(v=v, r=r, s=s)

    return SubmitSignatures(
        type=0,
        message=message,
        signature=signature,
        unsignedMessage=unsignedMessage,
    )


def submit_signatures_type_1(payload: bytes) -> SubmitSignatures:
    payload_bp = ByteParser(payload)
    signature_to_parse = payload_bp.next_n(65)
    unsignedMessage = payload_bp.drain()

    signature_bp = ByteParser(signature_to_parse)
    v = signature_bp.next_n(1).hex()
    r = signature_bp.next_n(32).hex()
    s = signature_bp.next_n(32).hex()

    signature = Signature(v=v, r=r, s=s)

    return SubmitSignatures(
        type=1,
        message=None,
        signature=signature,
        unsignedMessage=unsignedMessage,
    )


def submit_signatures(payload: bytes) -> SubmitSignatures:
    payload_bp = ByteParser(payload)
    type = payload_bp.uint8()
    rest_of_payload = payload_bp.drain()

    match type:
        case 0:
            return submit_signatures_type_0(rest_of_payload)
        case 1:
            return submit_signatures_type_1(rest_of_payload)
        case _:
            raise Exception(
                f"Version {type} of SubmitSignatures payload is not defined."
            )



__all__ = ["encode_update_array"]


def encode_update_array(deltas: list[int]) -> list[int]:
    update_array = []

    for number in deltas:
        if not (0 <= number <= 255):
            raise ValueError("The number must be between 0 and 255, inclusive.")
        for shift in range(7, -1, -2):
            sign_bit = (number >> shift) & 1
            value_bit = (number >> (shift - 1)) & 1
            update_array.append(-value_bit if sign_bit == 1 else value_bit)

    return update_array



from .parse import (
    parse_generic_tx,
    parse_submit1_tx,
    parse_submit2_tx,
    parse_submit_signature_tx,
)

__all__ = [
    "parse_generic_tx",
    "parse_submit1_tx",
    "parse_submit2_tx",
    "parse_submit_signature_tx",
]



import string

__all__ = []


def un_prefix_0x(s: str) -> str:
    return s.removeprefix("0x")


def prefix_0x(s: str) -> str:
    return "0x" + un_prefix_0x(s)


def is_hex_str(s: str):
    s = un_prefix_0x(s)
    return all(c in string.hexdigits for c in s)


def to_bytes(s: str | bytes) -> bytes:
    if isinstance(s, str):
        if not is_hex_str(s):
            raise ValueError("Invalid hex string")
        s = bytes.fromhex(s)
    return s



from functools import total_ordering
from typing import TYPE_CHECKING, Self

from attrs import frozen

if TYPE_CHECKING:
    from .factory import (
        Factory,
        RewardEpochFactory,
        VotingEpochFactory,
    )


__all__ = [
    "Epoch",
    "VotingEpoch",
    "RewardEpoch",
]


@total_ordering
@frozen
class Epoch:
    id: int
    factory: "Factory"

    def __lt__(self, other: Self) -> bool:
        if not isinstance(other, self.__class__) or self.factory != other.factory:
            return NotImplemented
        return self.id < other.id

    def __contains__(self, time: int):
        return self.start_s <= time < self.end_s

    @property
    def next(self) -> Self:
        return type(self)(self.id + 1, self.factory)

    @property
    def previous(self) -> Self:
        return type(self)(self.id - 1, self.factory)

    @property
    def start_s(self) -> int:
        return self.factory.first_epoch_epoc + self.id * self.factory.epoch_duration

    @property
    def end_s(self) -> int:
        return self.next.start_s


@frozen
class VotingEpoch(Epoch):
    factory: "VotingEpochFactory"

    def to_reward_epoch(self) -> "RewardEpoch":
        return self.factory.make_reward_epoch(self.start_s)

    def reveal_deadline(self) -> int:
        return self.start_s + self.factory.ftso_reveal_deadline


@frozen
class RewardEpoch(Epoch):
    factory: "RewardEpochFactory"

    def to_first_voting_epoch(self) -> VotingEpoch:
        return self.factory.make_voting_epoch(self.start_s)



from attrs import frozen

__all__ = ["calculate_median", "FtsoVote"]


@frozen
class FtsoVote:
    value: int
    weight: int


@frozen
class FtsoMedian:
    value: int
    first_quartile: int
    third_quartile: int
    sorted_votes: list[FtsoVote]


def calculate_median(votes: list[FtsoVote]) -> FtsoMedian | None:
    if len(votes) == 0:
        return None

    median = None
    quartile_1 = None
    quartile_3 = None

    # Sort the list by values if it is not already sorted.
    if not (all(votes[i].value <= votes[i + 1].value for i in range(len(votes) - 1))):
        votes.sort(key=lambda x: x.value)

    total_weight = sum([vote.weight for vote in votes])
    median_weight = total_weight // 2 + (total_weight % 2)
    quartile_weight = total_weight // 4
    current_weight_sum = 0

    for i, vote in enumerate(votes):
        current_weight_sum += vote.weight

        if quartile_1 is None and current_weight_sum > quartile_weight:
            quartile_1 = vote.value

        if current_weight_sum >= median_weight:
            if current_weight_sum == median_weight and total_weight % 2 == 0:
                next_vote = votes[i + 1]
                median = (vote.value + next_vote.value) // 2
                break
            else:
                median = vote.value
                break

    current_weight_sum = 0

    for i in range(len(votes) - 1, -1, -1):
        vote = votes[i]
        current_weight_sum += vote.weight

        if current_weight_sum > quartile_weight:
            quartile_3 = vote.value
            break

    assert median is not None
    assert quartile_1 is not None
    assert quartile_3 is not None

    return FtsoMedian(
        value=median,
        first_quartile=quartile_1,
        third_quartile=quartile_3,
        sorted_votes=votes,
    )



__all__ = []


class ParseError(Exception):
    pass


class ByteParser:
    def __init__(self, b: bytes) -> None:
        self._b = b
        self._pointer = 0

    def _move_pointer(self, n: int) -> int:
        p = self._pointer
        self._pointer += n
        if self._pointer > len(self._b):
            raise ParseError("Tried to parse bytes out of range.")
        return p

    def _consume(self, n: int) -> bytes:
        p = self._move_pointer(n)
        return self._b[p : self._pointer]

    def _parse_int(self, size: int, signed: bool) -> int:
        return int.from_bytes(self._consume(size), signed=signed)

    def uint8(self) -> int:
        return self._parse_int(1, False)

    def int8(self) -> int:
        return self._parse_int(1, True)

    def uint16(self) -> int:
        return self._parse_int(2, False)

    def int16(self) -> int:
        return self._parse_int(2, True)

    def uint32(self) -> int:
        return self._parse_int(4, False)

    def int32(self) -> int:
        return self._parse_int(4, True)

    def uint64(self) -> int:
        return self._parse_int(8, False)

    def int64(self) -> int:
        return self._parse_int(8, True)

    def uint128(self) -> int:
        return self._parse_int(16, False)

    def int128(self) -> int:
        return self._parse_int(16, True)

    def uint256(self) -> int:
        return self._parse_int(32, False)

    def int256(self) -> int:
        return self._parse_int(32, True)

    def next_n(self, n) -> bytes:
        return self._consume(n)

    def is_empty(self) -> bool:
        return self._pointer == len(self._b)

    def drain(self) -> bytes:
        return self._consume(len(self) - self._pointer)

    def __len__(self) -> int:
        return len(self._b)



from .commit import commit_hash
from .fast_updates import encode_update_array
from .feed import FtsoFeed
from .median import FtsoVote, calculate_median

__all__ = [
    "FtsoFeed",
    "calculate_median",
    "FtsoVote",
    "commit_hash",
    "encode_update_array",
]



import time

from attrs import frozen

from .epoch import Epoch, RewardEpoch, VotingEpoch

__all__ = [
    "Factory",
    "VotingEpochFactory",
    "RewardEpochFactory",
]


@frozen
class Factory:
    first_epoch_epoc: int
    epoch_duration: int

    def make_epoch(self, id) -> Epoch:
        return Epoch(id, self)

    def duration(self) -> int:
        return self.epoch_duration

    def _from_timestamp(self, ts: int) -> int:
        return (ts - self.first_epoch_epoc) // self.epoch_duration

    def from_timestamp(self, ts: int) -> Epoch:
        return self.make_epoch(self._from_timestamp(ts))

    def now(self) -> Epoch:
        return self.from_timestamp(int(time.time()))

    def now_id(self) -> int:
        return self.now().id


@frozen
class VotingEpochFactory(Factory):
    ftso_reveal_deadline: int
    # Reward Epoch data
    reward_first_epoch_epoc: int
    reward_epoch_duration: int

    def make_epoch(self, id) -> VotingEpoch:
        return VotingEpoch(id, self)

    def make_reward_epoch(self, t: int):
        factory = RewardEpochFactory(
            self.reward_first_epoch_epoc,
            self.reward_epoch_duration,
            self.first_epoch_epoc,
            self.epoch_duration,
            self.ftso_reveal_deadline,
        )
        id = factory._from_timestamp(t)
        return factory.make_epoch(id)


@frozen
class RewardEpochFactory(Factory):
    # Voting Epoch data
    voting_first_epoch_epoc: int
    voting_epoch_duration: int
    voting_ftso_reveal_deadline: int

    def make_epoch(self, id) -> RewardEpoch:
        return RewardEpoch(id, self)

    def make_voting_epoch(self, t: int):
        factory = VotingEpochFactory(
            self.voting_first_epoch_epoc,
            self.voting_epoch_duration,
            self.voting_ftso_reveal_deadline,
            self.first_epoch_epoc,
            self.epoch_duration,
        )
        id = factory._from_timestamp(t)
        return factory.make_epoch(id)

    def from_voting_epoch(self, voting_epoch: VotingEpoch) -> RewardEpoch:
        if not (
            voting_epoch.factory.first_epoch_epoc == self.voting_first_epoch_epoc
            and voting_epoch.factory.epoch_duration == self.voting_epoch_duration
            and voting_epoch.factory.ftso_reveal_deadline
            == self.voting_ftso_reveal_deadline
            and voting_epoch.factory.reward_first_epoch_epoc == self.first_epoch_epoc
            and voting_epoch.factory.reward_epoch_duration == self.epoch_duration
        ):
            raise ValueError("VotingEpoch was made by wrong factory")
        return self.make_epoch(self._from_timestamp(voting_epoch.start_s))



from hashlib import sha256

from base58 import BITCOIN_ALPHABET, b58decode, b58encode, scrub_input

__all__ = [
    "flare_b58_encode_check",
    "flare_b58_decode_check",
]


def flare_b58_encode_check(
    v: str | bytes,
    alphabet: bytes = BITCOIN_ALPHABET,
) -> bytes:
    """
    Encode a string using Base58 with a 4 character checksum like in avalanche
    and not like in spec
    https://github.com/flare-foundation/go-flare/blob/93fd844b1e85366ee9c1c4a3fb9e9399220534cc/avalanchego/utils/hashing/hashing.go#L78-L81
    """
    v = scrub_input(v)

    digest = sha256(v).digest()
    return b58encode(v + digest[-4:], alphabet=alphabet)


def flare_b58_decode_check(
    v: str | bytes, alphabet: bytes = BITCOIN_ALPHABET, *, autofix: bool = False
) -> bytes:
    """
    Encode a string using Base58 with a 4 character checksum like in avalanche
    and not like in spec
    https://github.com/flare-foundation/go-flare/blob/93fd844b1e85366ee9c1c4a3fb9e9399220534cc/avalanchego/utils/hashing/hashing.go#L78-L81
    """
    result = b58decode(v, alphabet=alphabet, autofix=autofix)
    result, check = result[:-4], result[-4:]
    digest = sha256(result).digest()

    if check != digest[-4:]:
        raise ValueError("Invalid checksum")

    return result



from attrs import Attribute, field, frozen

__all__ = ["FtsoFeed"]


@frozen
class FtsoFeed:
    @staticmethod
    def _length_validation(instance: "FtsoFeed", attribute: Attribute, value: bytes):
        if len(value) != 21:
            raise ValueError(f"{attribute.name} must have exactly 21 bytes.")

    feed_id: bytes = field(validator=_length_validation)

    @property
    def representation(self) -> str:
        return self.feed_id[1:].decode("utf-8").rstrip("\x00").strip()

    @property
    def type(self) -> int:
        return int(hex(self.feed_id[0]), 16)

    @classmethod
    def from_represenation_and_type(cls, type: int, representaion: str):
        encoded_type = type.to_bytes()
        encoded_rep = representaion.encode().ljust(20, b"\x00")

        return cls(encoded_type + encoded_rep)

    @classmethod
    def fromhex(cls, hexstr: str):
        return cls(bytes.fromhex(hexstr))



from eth_abi.abi import encode
from eth_hash.auto import keccak

__all__ = ["commit_hash"]


def commit_hash(
    submit_address: str, voting_round: int, random: int, feed_values: bytes
) -> str:
    types = ["address", "uint32", "uint256", "bytes"]
    values = [submit_address, voting_round, random, feed_values]
    encoded = encode(types, values)
    return keccak(encoded).hex()



from ...epoch import RewardEpoch, VotingEpoch
from ...factory import RewardEpochFactory, VotingEpochFactory
from ..config import coston2_chain_config

__all__ = [
    "voting_epoch",
    "reward_epoch",
    "voting_epoch_factory",
    "reward_epoch_factory",
]

voting_epoch_factory = VotingEpochFactory(
    first_epoch_epoc=coston2_chain_config.voting_first_epoch_epoc,
    epoch_duration=coston2_chain_config.voting_epoch_duration,
    ftso_reveal_deadline=coston2_chain_config.voting_ftso_reveal_deadline,
    reward_first_epoch_epoc=coston2_chain_config.reward_first_epoch_epoc,
    reward_epoch_duration=coston2_chain_config.reward_epoch_duration,
)

reward_epoch_factory = RewardEpochFactory(
    first_epoch_epoc=coston2_chain_config.reward_first_epoch_epoc,
    epoch_duration=coston2_chain_config.reward_epoch_duration,
    voting_first_epoch_epoc=coston2_chain_config.voting_first_epoch_epoc,
    voting_epoch_duration=coston2_chain_config.voting_epoch_duration,
    voting_ftso_reveal_deadline=coston2_chain_config.voting_ftso_reveal_deadline,
)


def voting_epoch(id: int) -> VotingEpoch:
    return voting_epoch_factory.make_epoch(id)


def reward_epoch(id: int) -> RewardEpoch:
    return reward_epoch_factory.make_epoch(id)



from eth_abi.abi import encode
from eth_hash.auto import keccak

from .._hexstr.hexstr import is_hex_str, prefix_0x, un_prefix_0x

__all__ = ["single_hash", "MerkleTree"]


def single_hash(value: str | bytes) -> str:
    if isinstance(value, str):
        if not is_hex_str(value):
            raise ValueError("Invalid hex string")
        value = bytes.fromhex(value)
    return prefix_0x(keccak(value).hex())


# Function to convert to hexadecimal and pad to bytes
def to_hex(value: str, pad_to_bytes: int) -> str:
    value = un_prefix_0x(value)
    padded_hex = value.rjust(pad_to_bytes * 2, "0")
    return "0x" + padded_hex


# Function to compute a sorted hash of two 32-byte strings
def sorted_hash_pair(x: str, y: str) -> str:
    x = un_prefix_0x(x)
    y = un_prefix_0x(y)
    xx = bytes.fromhex(x)
    yy = bytes.fromhex(y)
    if x <= y:
        encoded = encode(["bytes32", "bytes32"], [xx, yy])
    else:
        encoded = encode(["bytes32", "bytes32"], [yy, xx])
    return single_hash(encoded)


class MerkleTree:
    def __init__(self, values: list[str], initial_hash: bool = False):
        self._tree: list[str] = []
        self.initial_hash = initial_hash
        self.build(values)

    @property
    def root(self) -> str | None:
        return self._tree[0] if self._tree else None

    @property
    def root_bigint(self) -> int:
        return int(self.root, 16) if self.root else 0

    @property
    def tree(self) -> list[str]:
        return self._tree[:]

    @property
    def hash_count(self) -> int:
        return (len(self._tree) + 1) // 2

    @property
    def sorted_hashes(self) -> list[str]:
        return self._tree[self.hash_count - 1 :]

    def parent(self, i: int) -> int:
        return (i - 1) // 2

    def build(self, values: list[str]):
        sorted_values = sorted(to_hex(v, 32) for v in values)
        hashes = []
        for value in sorted_values:
            # To remove duplicates
            if not hashes or value != hashes[-1]:
                hashes.append(value)

        if self.initial_hash:
            hashes = [single_hash(h) for h in hashes]

        n = len(hashes)

        tree = [""] * (n - 1) + hashes

        i = n - 2
        while i >= 0:
            tree[i] = sorted_hash_pair(tree[2 * i + 1], tree[2 * i + 2])
            i -= 1
        self._tree = tree

    def get_hash(self, i: int) -> str | None:
        if self.hash_count == 0 or not (0 <= i < self.hash_count):
            return None
        pos = len(self._tree) - self.hash_count + i
        return self._tree[pos]

    def binary_search(self, target: str) -> int | None:
        low, high = 0, self.hash_count
        while high - low > 1:
            mid = (low + high) // 2
            if target < self.sorted_hashes[mid]:
                high = mid
            else:
                low = mid
        return low if self.sorted_hashes[low] == target else None

    def get_proof(self, target: str) -> list[str] | None:
        target = to_hex(target, 32)
        index = self.binary_search(target)
        if index is None:
            return None

        proof = []
        pos = len(self._tree) - self.hash_count + index
        while pos > 0:
            sibling = pos + 2 * (pos % 2) - 1
            proof.append(self._tree[sibling])
            pos = self.parent(pos)
        return proof


def verify_with_merkle_proof(leaf: str, proof: list[str], root: str) -> bool:
    if not leaf or not proof or not root:
        return False
    current_hash = leaf
    for sibling_hash in proof:
        current_hash = sorted_hash_pair(current_hash, sibling_hash)
    return current_hash == root



from .epoch import (
    reward_epoch,
    reward_epoch_factory,
    voting_epoch,
    voting_epoch_factory,
)

__all__ = [
    "voting_epoch",
    "reward_epoch",
    "reward_epoch_factory",
    "voting_epoch_factory",
]



from attrs import frozen

__all__ = []


@frozen
class ChainConfig:
    voting_first_epoch_epoc: int
    voting_epoch_duration: int
    voting_ftso_reveal_deadline: int
    reward_first_epoch_epoc: int
    reward_epoch_duration: int


# flare
flare_chain_config = ChainConfig(
    voting_first_epoch_epoc=1658430000,
    voting_epoch_duration=90,
    voting_ftso_reveal_deadline=45,
    reward_first_epoch_epoc=1658430000,
    reward_epoch_duration=302400,
)

# songbird
songbird_chain_config = ChainConfig(
    voting_first_epoch_epoc=1658429955,
    voting_epoch_duration=90,
    voting_ftso_reveal_deadline=45,
    reward_first_epoch_epoc=1658429955,
    reward_epoch_duration=302400,
)

# coston
coston_chain_config = ChainConfig(
    voting_first_epoch_epoc=1658429955,
    voting_epoch_duration=90,
    voting_ftso_reveal_deadline=45,
    reward_first_epoch_epoc=1658429955,
    reward_epoch_duration=21600,
)

# coston2
coston2_chain_config = ChainConfig(
    voting_first_epoch_epoc=1658430000,
    voting_epoch_duration=90,
    voting_ftso_reveal_deadline=45,
    reward_first_epoch_epoc=1658430000,
    reward_epoch_duration=21600,
)



from attrs import Attribute, field, frozen

__all__ = ["AttestationType"]


@frozen
class AttestationType:
    @staticmethod
    def _length_validation(
        instance: "AttestationType", attribute: Attribute, value: bytes
    ):
        if len(value) != 32:
            raise ValueError(f"{attribute.name} must have exactly 32 bytes.")

    attestation_type: bytes = field(validator=_length_validation)

    @property
    def representation(self) -> str:
        return self.attestation_type.decode("utf-8").rstrip("\x00").strip()

    @classmethod
    def from_represenation(cls, representaion: str):
        encoded_rep = representaion.encode().ljust(32, b"\x00")

        return cls(encoded_rep)

    @classmethod
    def fromhex(cls, hexstr: str):
        return cls(bytes.fromhex(hexstr))



from typing import Generic, TypeVar

from attrs import frozen

__all__ = []

T = TypeVar("T")
U = TypeVar("U")


@frozen
class ParsedPayload(Generic[T]):
    protocol_id: int
    voting_round_id: int
    size: int
    payload: T


@frozen
class ParsedMessage(Generic[T, U]):
    fdc: ParsedPayload[U] | None
    ftso: ParsedPayload[T] | None


@frozen
class FtsoSubmit1:
    commit_hash: bytes


@frozen
class FtsoSubmit2:
    random: int
    values: list[int | None]


@frozen
class FdcSubmit1:
    pass


@frozen
class FdcSubmit2:
    number_of_requests: int
    bit_vector: list[bool]


@frozen
class SubmitSignaturesMessage:
    protocol_id: int
    random_quality_score: int
    merkle_root: str


@frozen
class Signature:
    v: str
    r: str
    s: str


@frozen
class SubmitSignatures:
    type: int
    message: SubmitSignaturesMessage | None
    signature: Signature
    unsignedMessage: bytes



from .coston.epoch import reward_epoch as coston_reward_epoch
from .coston.epoch import reward_epoch_factory as coston2_reward_epoch_factory
from .coston.epoch import reward_epoch_factory as coston_reward_epoch_factory
from .coston.epoch import reward_epoch_factory as flare_reward_epoch_factory
from .coston.epoch import reward_epoch_factory as songbird_reward_epoch_factory
from .coston.epoch import voting_epoch as coston_voting_epoch
from .coston.epoch import voting_epoch_factory as coston2_voting_epoch_factory
from .coston.epoch import voting_epoch_factory as coston_voting_epoch_factory
from .coston.epoch import voting_epoch_factory as flare_voting_epoch_factory
from .coston.epoch import voting_epoch_factory as songbird_voting_epoch_factory
from .coston2.epoch import reward_epoch as coston2_reward_epoch
from .coston2.epoch import voting_epoch as coston2_voting_epoch
from .flare.epoch import reward_epoch as flare_reward_epoch
from .flare.epoch import voting_epoch as flare_voting_epoch
from .songbird.epoch import reward_epoch as songbird_reward_epoch
from .songbird.epoch import voting_epoch as songbird_voting_epoch

__all__ = [
    "coston_voting_epoch",
    "coston_reward_epoch",
    "coston_voting_epoch_factory",
    "coston_reward_epoch_factory",
    "coston2_voting_epoch",
    "coston2_reward_epoch",
    "coston2_voting_epoch_factory",
    "coston2_reward_epoch_factory",
    "songbird_voting_epoch",
    "songbird_reward_epoch",
    "songbird_voting_epoch_factory",
    "songbird_reward_epoch_factory",
    "flare_voting_epoch",
    "flare_reward_epoch",
    "flare_voting_epoch_factory",
    "flare_reward_epoch_factory",
]



from .epoch import RewardEpoch, VotingEpoch
from .factory import RewardEpochFactory, VotingEpochFactory

__all__ = [
    "RewardEpoch",
    "VotingEpoch",
    "VotingEpochFactory",
    "RewardEpochFactory",
]
